@@ -18,6 +18,7 @@
 #define __PYTHON_H__
 
 #include "../common/py_version.hpp"
+#include <cstdint>
 
 #ifndef _WIN32
 typedef unsigned int DWORD;
@@ -48,217 +49,6 @@ public:
     size_t ob_size; /* Number of items in variable part */
 };
 
-// 2.4 - 2.7 compatible
-class PyCodeObject25_27 : public PyObject {
-public:
-    int co_argcount;        /* #arguments, except *args */
-    int co_nlocals;         /* #local variables */
-    int co_stacksize;       /* #entries needed for evaluation stack */
-    int co_flags;           /* CO_..., see below */
-    PyObject *co_code;      /* instruction opcodes */
-    PyObject *co_consts;    /* list (constants used) */
-    PyObject *co_names;     /* list of strings (names used) */
-    PyObject *co_varnames;  /* tuple of strings (local variable names) */
-    PyObject *co_freevars;  /* tuple of strings (free variable names) */
-    PyObject *co_cellvars;  /* tuple of strings (cell variable names) */
-    /* The rest doesn't count for hash/cmp */
-    PyObject *co_filename;  /* string (where it was loaded from) */
-    PyObject *co_name;      /* string (name, for reference) */
-    int co_firstlineno;     /* first source line number */
-    PyObject *co_lnotab;    /* string (encoding addr<->lineno mapping) */
-
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 2 && (minorVersion >= 5 && minorVersion <= 7);
-    }
-
-    static bool IsFor(PythonVersion version) {
-        return version >= PythonVersion_25 && version <= PythonVersion_27;
-    }
-};
-
-// 3.0-3.2
-class PyCodeObject30_32 : public PyObject {
-public:
-    int co_argcount;        /* #arguments, except *args */
-    int co_kwonlyargcount;  /* #keyword only arguments */
-    int co_nlocals;         /* #local variables */
-    int co_stacksize;       /* #entries needed for evaluation stack */
-    int co_flags;           /* CO_..., see below */
-    PyObject *co_code;      /* instruction opcodes */
-    PyObject *co_consts;    /* list (constants used) */
-    PyObject *co_names;     /* list of strings (names used) */
-    PyObject *co_varnames;  /* tuple of strings (local variable names) */
-    PyObject *co_freevars;  /* tuple of strings (free variable names) */
-    PyObject *co_cellvars;  /* tuple of strings (cell variable names) */
-    /* The rest doesn't count for hash or comparisons */
-    PyObject *co_filename;  /* unicode (where it was loaded from) */
-    PyObject *co_name;      /* unicode (name, for reference) */
-    int co_firstlineno;     /* first source line number */
-    PyObject *co_lnotab;    /* string (encoding addr<->lineno mapping) */
-    void *co_zombieframe;   /* for optimization only (see frameobject.c) */
-
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 3 && (minorVersion >= 0 && minorVersion <= 2);
-    }
-
-    static bool IsFor(PythonVersion version) {
-        return version >= PythonVersion_30 && version <= PythonVersion_32;
-    }
-};
-
-// 3.3-3.5
-class PyCodeObject33_35 : public PyObject {
-public:
-    int co_argcount;            /* #arguments, except *args */
-    int co_kwonlyargcount;      /* #keyword only arguments */
-    int co_nlocals;             /* #local variables */
-    int co_stacksize;           /* #entries needed for evaluation stack */
-    int co_flags;               /* CO_..., see below */
-    PyObject *co_code;          /* instruction opcodes */
-    PyObject *co_consts;        /* list (constants used) */
-    PyObject *co_names;         /* list of strings (names used) */
-    PyObject *co_varnames;      /* tuple of strings (local variable names) */
-    PyObject *co_freevars;      /* tuple of strings (free variable names) */
-    PyObject *co_cellvars;      /* tuple of strings (cell variable names) */
-    /* The rest doesn't count for hash or comparisons */
-    unsigned char *co_cell2arg; /* Maps cell vars which are arguments. */
-    PyObject *co_filename;      /* unicode (where it was loaded from) */
-    PyObject *co_name;          /* unicode (name, for reference) */
-    int co_firstlineno;         /* first source line number */
-    PyObject *co_lnotab;        /* string (encoding addr<->lineno mapping) */
-    void *co_zombieframe;       /* for optimization only (see frameobject.c) */
-
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 3 && (minorVersion >= 3 && minorVersion <= 5);
-    }
-
-    static bool IsFor(PythonVersion version) {
-        return version >= PythonVersion_33 && version <= PythonVersion_35;
-    }
-};
-
-// 3.6
-class PyCodeObject36 : public PyObject {
-public:
-    int co_argcount;            /* #arguments, except *args */
-    int co_kwonlyargcount;      /* #keyword only arguments */
-    int co_nlocals;             /* #local variables */
-    int co_stacksize;           /* #entries needed for evaluation stack */
-    int co_flags;               /* CO_..., see below */
-    int co_firstlineno;         /* first source line number */
-    PyObject *co_code;          /* instruction opcodes */
-    PyObject *co_consts;        /* list (constants used) */
-    PyObject *co_names;         /* list of strings (names used) */
-    PyObject *co_varnames;      /* tuple of strings (local variable names) */
-    PyObject *co_freevars;      /* tuple of strings (free variable names) */
-    PyObject *co_cellvars;      /* tuple of strings (cell variable names) */
-    /* The rest doesn't count for hash or comparisons */
-    unsigned char *co_cell2arg; /* Maps cell vars which are arguments. */
-    PyObject *co_filename;      /* unicode (where it was loaded from) */
-    PyObject *co_name;          /* unicode (name, for reference) */
-    PyObject *co_lnotab;        /* string (encoding addr<->lineno mapping) */
-    void *co_zombieframe;       /* for optimization only (see frameobject.c) */
-
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 3 && minorVersion == 6;
-    }
-
-    static bool IsFor(PythonVersion version) {
-        return version == PythonVersion_36;
-    }
-};
-
-// 3.7
-class PyCodeObject37 : public PyObject {
-public:
-    int co_argcount;            /* #arguments, except *args */
-    int co_kwonlyargcount;      /* #keyword only arguments */
-    int co_nlocals;             /* #local variables */
-    int co_stacksize;           /* #entries needed for evaluation stack */
-    int co_flags;               /* CO_..., see below */
-    int co_firstlineno;         /* first source line number */
-    PyObject *co_code;          /* instruction opcodes */
-    PyObject *co_consts;        /* list (constants used) */
-    PyObject *co_names;         /* list of strings (names used) */
-    PyObject *co_varnames;      /* tuple of strings (local variable names) */
-    PyObject *co_freevars;      /* tuple of strings (free variable names) */
-    PyObject *co_cellvars;      /* tuple of strings (cell variable names) */
-                                /* The rest doesn't count for hash or comparisons */
-    SSIZE_T *co_cell2arg;       /* Maps cell vars which are arguments. */
-    PyObject *co_filename;      /* unicode (where it was loaded from) */
-    PyObject *co_name;          /* unicode (name, for reference) */
-    PyObject *co_lnotab;        /* string (encoding addr<->lineno mapping) */
-    void *co_zombieframe;       /* for optimization only (see frameobject.c) */
-
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 3 && minorVersion == 7;
-    }
-
-    static bool IsFor(PythonVersion version) {
-        return version == PythonVersion_37;
-    }
-};
-
-typedef struct _PyOpcache _PyOpcache;
-
-// 3.8
-class PyCodeObject38 : public PyObject {
-public:
-    int co_argcount;            /* #arguments, except *args */
-    int co_posonlyargcount;     /* #positional only arguments */
-    int co_kwonlyargcount;      /* #keyword only arguments */
-    int co_nlocals;             /* #local variables */
-    int co_stacksize;           /* #entries needed for evaluation stack */
-    int co_flags;               /* CO_..., see below */
-    int co_firstlineno;         /* first source line number */
-    PyObject *co_code;          /* instruction opcodes */
-    PyObject *co_consts;        /* list (constants used) */
-    PyObject *co_names;         /* list of strings (names used) */
-    PyObject *co_varnames;      /* tuple of strings (local variable names) */
-    PyObject *co_freevars;      /* tuple of strings (free variable names) */
-    PyObject *co_cellvars;      /* tuple of strings (cell variable names) */
-    /* The rest aren't used in either hash or comparisons, except for co_name,
-       used in both. This is done to preserve the name and line number
-       for tracebacks and debuggers; otherwise, constant de-duplication
-       would collapse identical functions/lambdas defined on different lines.
-    */
-    SSIZE_T *co_cell2arg;    /* Maps cell vars which are arguments. */
-    PyObject *co_filename;      /* unicode (where it was loaded from) */
-    PyObject *co_name;          /* unicode (name, for reference) */
-    PyObject *co_lnotab;        /* string (encoding addr<->lineno mapping) See
-                                   Objects/lnotab_notes.txt for details. */
-    void *co_zombieframe;       /* for optimization only (see frameobject.c) */
-    PyObject *co_weakreflist;   /* to support weakrefs to code objects */
-    /* Scratch space for extra data relating to the code object.
-       Type is a void* to keep the format private in codeobject.c to force
-       people to go through the proper APIs. */
-    void *co_extra;
-
-    /* Per opcodes just-in-time cache
-     *
-     * To reduce cache size, we use indirect mapping from opcode index to
-     * cache object:
-     *   cache = co_opcache[co_opcache_map[next_instr - first_instr] - 1]
-     */
-
-    // co_opcache_map is indexed by (next_instr - first_instr).
-    //  * 0 means there is no cache for this opcode.
-    //  * n > 0 means there is cache in co_opcache[n-1].
-    unsigned char *co_opcache_map;
-    _PyOpcache *co_opcache;
-    int co_opcache_flag;  // used to determine when create a cache.
-    unsigned char co_opcache_size;  // length of co_opcache.
-
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 3 && minorVersion == 8;
-    }
-
-    static bool IsFor(PythonVersion version) {
-        return version == PythonVersion_38;
-    }
-
-};
-
 // 2.5 - 3.7
 class PyFunctionObject : public PyObject {
 public:
@@ -290,10 +80,20 @@ typedef struct {
     long hash;          /* Hash value; -1 if not set */
 } PyUnicodeObject;
 
+
+class PyFrameObject : public PyObject {
+    // After 3.10 we don't really have things we want to reuse common, so,
+    // create an empty base (it's not based on PyVarObject because on Python 3.11
+    // it's just a PyObject and no longer a PyVarObject -- the part related to
+    // the var object must be declared in ech subclass in this case).
+};
+
 // 2.4 - 3.7 compatible
-class PyFrameObject : public PyVarObject {
+class PyFrameObjectBaseUpTo39 : public PyFrameObject {
 public:
-    PyFrameObject *f_back;  /* previous frame, or nullptr */
+    size_t ob_size; /* Number of items in variable part -- i.e.: PyVarObject*/
+    
+    PyFrameObjectBaseUpTo39 *f_back;  /* previous frame, or nullptr */
     PyObject *f_code;           /* code segment */
     PyObject *f_builtins;       /* builtin symbol table (PyDictObject) */
     PyObject *f_globals;        /* global symbol table (PyDictObject) */
@@ -304,72 +104,65 @@ public:
        to the current stack top. */
     PyObject **f_stacktop;
     PyObject *f_trace;          /* Trace function */
+    
+    // It has more things, but we're only interested in things up to f_trace.
+
 };
 
-#define CO_MAXBLOCKS 20
-typedef struct {
-    int b_type;         /* what kind of block this is */
-    int b_handler;      /* where to jump to find handler */
-    int b_level;        /* value stack level to pop to */
-} PyTryBlock;
 
-class PyFrameObject25_33 : public PyFrameObject {
+// https://github.com/python/cpython/blob/3.10/Include/cpython/frameobject.h
+class PyFrameObject310 : public PyFrameObject {
 public:
-    PyObject * f_exc_type, *f_exc_value, *f_exc_traceback;
-    PyThreadState* f_tstate;
-    int f_lasti;                /* Last instruction if called */
-    /* As of 2.3 f_lineno is only valid when tracing is active (i.e. when
-       f_trace is set) -- at other times use PyCode_Addr2Line instead. */
-    int f_lineno;               /* Current line number */
-    int f_iblock;       /* index in f_blockstack */
-    PyTryBlock f_blockstack[CO_MAXBLOCKS]; /* for try and loop blocks */
-    PyObject *f_localsplus[1];    /* locals+stack, dynamically sized */
+    size_t ob_size; /* Number of items in variable part -- i.e.: PyVarObject*/
 
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return (majorVersion == 2 && (minorVersion >= 5 && minorVersion <= 7)) ||
-            (majorVersion == 3 && (minorVersion >= 0 && minorVersion <= 3));
-    }
+    PyFrameObject310 *f_back;      /* previous frame, or NULL */
+    PyObject *f_code;       /* code segment */
+    PyObject *f_builtins;       /* builtin symbol table (PyDictObject) */
+    PyObject *f_globals;        /* global symbol table (PyDictObject) */
+    PyObject *f_locals;         /* local symbol table (any mapping) */
+    PyObject **f_valuestack;    /* points after the last local */
+    PyObject *f_trace;          /* Trace function */
+
+    // It has more things, but we're only interested in things up to f_trace.
 };
 
-class PyFrameObject34_36 : public PyFrameObject {
-public:
-    PyObject * f_exc_type, *f_exc_value, *f_exc_traceback;
-    /* Borrowed reference to a generator, or nullptr */
-    PyObject *f_gen;
+typedef uint16_t _Py_CODEUNIT;
 
-    int f_lasti;                /* Last instruction if called */
-    /* As of 2.3 f_lineno is only valid when tracing is active (i.e. when
-       f_trace is set) -- at other times use PyCode_Addr2Line instead. */
-    int f_lineno;               /* Current line number */
-    int f_iblock;       /* index in f_blockstack */
-    char f_executing;           /* whether the frame is still executing */
-    PyTryBlock f_blockstack[CO_MAXBLOCKS]; /* for try and loop blocks */
-    PyObject *f_localsplus[1];    /* locals+stack, dynamically sized */
+// https://github.com/python/cpython/blob/3.11/Include/internal/pycore_frame.h
+typedef struct _PyInterpreterFrame311 {
+    /* "Specials" section */
+    PyFunctionObject *f_func; /* Strong reference */
+    PyObject *f_globals; /* Borrowed reference */
+    PyObject *f_builtins; /* Borrowed reference */
+    PyObject *f_locals; /* Strong reference, may be NULL */
+    void *f_code; /* Strong reference */
+    void *frame_obj; /* Strong reference, may be NULL */
+    /* Linkage section */
+    struct _PyInterpreterFrame311 *previous;
+    // NOTE: This is not necessarily the last instruction started in the given
+    // frame. Rather, it is the code unit *prior to* the *next* instruction. For
+    // example, it may be an inline CACHE entry, an instruction we just jumped
+    // over, or (in the case of a newly-created frame) a totally invalid value:
+    _Py_CODEUNIT *prev_instr;
+    int stacktop;     /* Offset of TOS from localsplus  */
+    bool is_entry;  // Whether this is the "root" frame for the current _PyCFrame.
+    char owner;
+    /* Locals and stack */
+    PyObject *localsplus[1];
+} _PyInterpreterFrame311;
 
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 3 && minorVersion >= 4 && minorVersion <= 6;
-    }
-};
-
-class PyFrameObject37_38 : public PyFrameObject {
+// https://github.com/python/cpython/blob/3.11/Include/internal/pycore_frame.h
+// Note that in 3.11 it's no longer a "PyVarObject".
+class PyFrameObject311 : public PyFrameObject {
 public:
+    PyFrameObject311 *f_back;      /* previous frame, or NULL */
+    struct _PyInterpreterFrame311 *f_frame; /* points to the frame data */
+    PyObject *f_trace;          /* Trace function */
+    int f_lineno;               /* Current line number. Only valid if non-zero */
     char f_trace_lines;         /* Emit per-line trace events? */
     char f_trace_opcodes;       /* Emit per-opcode trace events? */
-    /* Borrowed reference to a generator, or nullptr */
-    PyObject *f_gen;
-
-    int f_lasti;                /* Last instruction if called */
-    /* As of 2.3 f_lineno is only valid when tracing is active (i.e. when
-       f_trace is set) -- at other times use PyCode_Addr2Line instead. */
-    int f_lineno;               /* Current line number */
-    int f_iblock;       /* index in f_blockstack */
-    char f_executing;           /* whether the frame is still executing */
-    PyTryBlock f_blockstack[CO_MAXBLOCKS]; /* for try and loop blocks */
-    PyObject *f_localsplus[1];    /* locals+stack, dynamically sized */
-
-    static bool IsFor(int majorVersion, int minorVersion) {
-        return majorVersion == 3 && minorVersion >= 7;
-    }
+    char f_fast_as_locals;      /* Have the fast locals of this frame been converted to a dict? */
+    // It has more things, but we're not interested on those.
 };
 
 
@@ -509,7 +302,7 @@ public:
     PyThreadState *next;
     PyInterpreterState *interp;
 
-    PyFrameObject *frame;
+    PyFrameObjectBaseUpTo39 *frame;
     int recursion_depth;
     /* 'tracing' keeps track of the execution depth when tracing/profiling.
        This is to prevent the actual trace/profile code from being recorded in
@@ -560,7 +353,7 @@ public:
     PyThreadState *next;
     PyInterpreterState *interp;
 
-    PyFrameObject *frame;
+    PyFrameObjectBaseUpTo39 *frame;
     int recursion_depth;
     char overflowed; /* The stack has overflowed. Allow 50 more calls
                         to handle the runtime error. */
@@ -616,7 +409,7 @@ public:
     PyThreadState *next;
     PyInterpreterState *interp;
 
-    PyFrameObject *frame;
+    PyFrameObjectBaseUpTo39 *frame;
     int recursion_depth;
     char overflowed; /* The stack has overflowed. Allow 50 more calls
                         to handle the runtime error. */
@@ -671,7 +464,7 @@ public:
     PyThreadState *next;
     PyInterpreterState *interp;
 
-    PyFrameObject *frame;
+    PyFrameObjectBaseUpTo39 *frame;
     int recursion_depth;
     char overflowed; /* The stack has overflowed. Allow 50 more calls
                      to handle the runtime error. */
@@ -721,7 +514,7 @@ public:
     PyThreadState *next;
     PyInterpreterState *interp;
 
-    PyFrameObject *frame;
+    PyFrameObjectBaseUpTo39 *frame;
     int recursion_depth;
     char overflowed; /* The stack has overflowed. Allow 50 more calls
                      to handle the runtime error. */
@@ -759,6 +552,126 @@ public:
     }
 };
 
+typedef struct _cframe {
+    /* This struct will be threaded through the C stack
+     * allowing fast access to per-thread state that needs
+     * to be accessed quickly by the interpreter, but can
+     * be modified outside of the interpreter.
+     *
+     * WARNING: This makes data on the C stack accessible from
+     * heap objects. Care must be taken to maintain stack
+     * discipline and make sure that instances of this struct cannot
+     * accessed outside of their lifetime.
+     */
+    int use_tracing;
+    struct _cframe *previous;
+} CFrame;
+
+// i.e.: https://github.com/python/cpython/blob/master/Include/cpython/pystate.h
+class PyThreadState_310 : public PyThreadState {
+public:
+    PyThreadState *prev;
+    PyThreadState *next;
+    PyInterpreterState *interp;
+
+    PyFrameObject310 *frame;
+    int recursion_depth;
+    int recursion_headroom; /* Allow 50 more calls to handle any errors. */
+    int stackcheck_counter;
+
+    /* 'tracing' keeps track of the execution depth when tracing/profiling.
+       This is to prevent the actual trace/profile code from being recorded in
+       the trace/profile. */
+    int tracing;
+
+    /* Pointer to current CFrame in the C stack frame of the currently,
+     * or most recently, executing _PyEval_EvalFrameDefault. */
+    CFrame *cframe;
+
+
+    Py_tracefunc c_profilefunc;
+    Py_tracefunc c_tracefunc;
+    PyObject *c_profileobj;
+    PyObject *c_traceobj;
+
+    PyObject *curexc_type;
+    PyObject *curexc_value;
+    PyObject *curexc_traceback;
+
+    _PyErr_StackItem exc_state;
+    _PyErr_StackItem *exc_info;
+
+    PyObject *dict;  /* Stores per-thread state */
+
+    int gilstate_counter;
+
+    PyObject *async_exc; /* Asynchronous exception to raise */
+
+    unsigned long thread_id; /* Thread id where this tstate was created */
+
+    static bool IsFor(int majorVersion, int minorVersion) {
+        return majorVersion == 3 && minorVersion == 10;
+    }
+
+    static bool IsFor(PythonVersion version) {
+        return version == PythonVersion_310;
+    }
+};
+
+// i.e.: https://github.com/python/cpython/blob/3.11/Include/cpython/pystate.h
+class PyThreadState_311 : public PyThreadState {
+public:
+    PyThreadState *prev;
+    PyThreadState *next;
+    PyInterpreterState *interp;
+
+    int _initialized;
+
+    int _static;
+    
+    int recursion_remaining;
+    int recursion_limit;
+    int recursion_headroom; /* Allow 50 more calls to handle any errors. */
+
+    /* 'tracing' keeps track of the execution depth when tracing/profiling.
+       This is to prevent the actual trace/profile code from being recorded in
+       the trace/profile. */
+    int tracing;
+    int tracing_what;
+
+    /* Pointer to current CFrame in the C stack frame of the currently,
+     * or most recently, executing _PyEval_EvalFrameDefault. */
+    CFrame *cframe;
+
+
+    Py_tracefunc c_profilefunc;
+    Py_tracefunc c_tracefunc;
+    PyObject *c_profileobj;
+    PyObject *c_traceobj;
+
+    PyObject *curexc_type;
+    PyObject *curexc_value;
+    PyObject *curexc_traceback;
+
+    _PyErr_StackItem *exc_info;
+
+    PyObject *dict;  /* Stores per-thread state */
+
+    int gilstate_counter;
+
+    PyObject *async_exc; /* Asynchronous exception to raise */
+
+    unsigned long thread_id; /* Thread id where this tstate was created */
+
+    static bool IsFor(int majorVersion, int minorVersion) {
+        return majorVersion == 3 && minorVersion == 11;
+    }
+
+    static bool IsFor(PythonVersion version) {
+        return version == PythonVersion_311;
+    }
+};
+
 class PyIntObject : public PyObject {
 public:
     long ob_ival;
